from asyncio import Queue

from ._task import Task


class TaskQueue(Queue):

    def __init__(self):
        super().__init__()

        self.queue = Queue()

    async def get(self) -> Task:
        return await self.queue.get()

    async def put(self, task: Task):
        await self.queue.put(task)



from ._worker import Worker
from ._task_queue import TaskQueue

class WorkerPool:

    def __init__(self, concurrency:int) -> None:
        """
        The main worker pool class.
        :param concurrency: The number of concurrency workers.
        """

        # initialize the worker queue
        _worker_queue = TaskQueue()


        # initialize the worker
        self._workers = [Worker(_worker_queue) for _ in range(concurrency)]

    def start(self):
        for worker in self._workers:
            worker.start()


from fastapi import FastAPI
from ._ping import ping_router


def create_rest_app() -> FastAPI:
    """
    Create and configure a FastAPI REST application.
    :return: The configured FastAPI REST application.
    """
    app = FastAPI(title="Ventricle")

    app.include_router(ping_router)

    return app


from typing import Optional

from fastapi import APIRouter
from pydantic import BaseModel

ping_router = APIRouter(prefix="/ping")

class PingResponse(BaseModel):
    status: int
    message: Optional[str] = None

@ping_router.get("", response_model=PingResponse)
async def ping_endpoint():
    return PingResponse(status=200)


import asyncio
import threading
from ._task_queue import TaskQueue


class Worker(threading.Thread):

    def __init__(self, queue: TaskQueue):
        """
        The Worker Object.
        :param queue: The queue to get the tasks from.
        """
        super().__init__()
        self.queue = queue

    def run(self):
        asyncio.run(self._run_async())

    async def _run_async(self):
        while True:
            task = await self.queue.get()
            try:
                await task.func(*task.args, **task.kwargs)
            except Exception as e:
                print(e) # todo replace with logging




from typing import Optional

import uvicorn

from .rest import create_rest_app
from .scheduler import create_schedular_app
from .worker import create_worker_app


class Ventricle:

    def __init__(self,
                 worker_concurrency: int = 16,
                 redis_url: Optional[str] = None,
                 amqp_url: Optional[str] = None):
        """
        The ventricle app.
        :param worker_concurrency: The number of worker processes to run in parallel.
        :param redis_url: The redis url to connect to. Some services may not work if not provided.
        :param amqp_url: The amqp url to connect to. Some services may not work if not provided.
        """
        self.rest = create_rest_app()
        self.scheduler = create_schedular_app()

        assert worker_concurrency > 0, "Worker concurrency must be greater than 0."
        self.worker = create_worker_app(worker_concurrency)

        # todo validate the urls
        self.REDIS_URL = redis_url
        self.AMQP_URL = amqp_url

    def start(self, rest: bool = True, schedular: bool = True, worker:bool=True) -> None:
        """
        Start the Ventricle app.
        :param rest: If the REST server should be started.
        :param schedular: If the scheduler should be started.
        :param worker: If the worker should be started.
        :return: None
        """
        if schedular:
            self.scheduler.start()

        if worker:
            self.worker.start()

        if rest:
            uvicorn.run(
                self.rest,
                host="0.0.0.0",
                port=8000
            )
